//
// nam_interface.cc : NAM Interface for Diffusion IO Log
// author           : Fabio Silva
//
// Copyright (C) 2000-2002 by the University of Southern California
// $Id: nam_interface.cc,v 1.8 2002/12/06 23:42:25 fabio Exp $
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License,
// version 2, as published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//
//

#include "nam_interface.hh"

char *msg_types[] = {"Interest", "Positive Reinforcement",
		     "Negative Reinforcement",
		     "Push Exploratory Data", "Exploratory Data", "Data"};

// Color Table for Packets
// 0 = black
// 1 = grey
// 2 = violet
// 3 = magenta
// 4 = yellow
// 5 = blue
// 6 = orange
// 7 = red
// 8 = green
// 9 = grey
// 10 = red2
// 11 = red3
// 12 = blue

// Set the following array according to the color table above
int msg_colors[] = {7, 6, 10, 12, 12, 5};

int SendDummyEventTimer::expire()
{
  return (app_->sendDummyEvent());
}

int SendLabelEventTimer::expire()
{
  return (app_->sendLabelEvent());
}

float NamInterface::getCurrentTime()
{
  float simulation_time;
  struct timeval now;

  // Calculate the 'current' time
  GetTime(&now);

  if (now.tv_usec < zero_.tv_usec){
    now.tv_usec = now.tv_usec + 1000000;
    now.tv_sec--;
  }

  simulation_time = now.tv_sec - zero_.tv_sec;
  simulation_time = simulation_time + (((float) (now.tv_usec - zero_.tv_usec)) / 1000000);

  return simulation_time;
}

int NamInterface::sendDummyEvent()
{
  float now = getCurrentTime();

  // Send a dummy event to nam to keep time moving forward
  printf("T -t %f\n", now);
  fflush(NULL);

  return 0;
}

int NamInterface::sendLabelEvent()
{
  float now = getCurrentTime();

  // Send a label event to the simulator
  printf("v -t %f -e sim_annotation %f %d Time: %f Packets: In:%d Out:%d\n",
	 now, now, labels_, now, packets_in_, packets_out_);
  fflush(NULL);
  labels_++;

  // Re-schedule timer so we are called again soon
  return 0;
}

void NamInterface::configNam()
{
  // Initialize simulation time
  GetTime(&zero_);

  // Print initial config
  printf("V -t * -v 1.0a11 -a 0\n");
  printf("W -t * -x %d -y %d\n", size_x_, size_y_);
  printf("A -t * -n 1 -p 0 -o 0xffffffff -c 31 -a 1\n");
  printf("A -t * -h 1 -m 2147483647 -s 0\n");
  printf("c -t * -i 0 -n black\n");
  printf("c -t * -i 1 -n grey\n");
  printf("c -t * -i 2 -n violet\n");
  printf("c -t * -i 3 -n magenta\n");
  printf("c -t * -i 4 -n yellow\n");
  printf("c -t * -i 5 -n blue\n");
  printf("c -t * -i 6 -n orange\n");
  printf("c -t * -i 7 -n red\n");
  printf("c -t * -i 8 -n green\n");
  printf("c -t * -i 9 -n grey\n");
  printf("c -t * -i 10 -n red2\n");
  printf("c -t * -i 11 -n red3\n");
  printf("c -t * -i 12 -n blue\n");
  printf("T -t 0.0\n");
  fflush(NULL);
}

void NamInterface::addNode(NodeInfo *node)
{
  float now;

  now = getCurrentTime();

  printf("n -t %f -s %d -v %s -c %s -z %d -x %d -y %d -S UP -w\n",
	 now, node->node_id_, node->node_shape_, node->node_color_,
	 node_size_, node->node_x_, node->node_y_);
  fflush(NULL);

  printf("n -t %f -s %d -S DLABEL -l \"out:0 in:0\" -L \"\"\n",
	 now, node->node_id_);
  fflush(NULL);

  node->initialized_ = true;
}

NodeInfo * NamInterface::findNode(int id)
{
  NodeList::iterator node_iterator;
  NodeInfo *node_entry;

  for (node_iterator = nodes_.begin(); node_iterator != nodes_.end(); ++node_iterator){
    node_entry = *node_iterator;

    if (node_entry->node_id_ == id)
      return node_entry;
  }

  // Node cannot be found !
  return NULL;
}

void NamInterface::sendPacket(int src, int dst)
{
  float now = getCurrentTime();

  printf("h -t %f -s %d -d %d -p AODV -e 200000 -c %d -a 0 -i 0 -k MAC\n",
	 now, src, dst, dst);
  fflush(NULL);
}

void NamInterface::processDiffusionLogEvent(int node_id, bool sending,
					    bool broadcast, int len, int extra,
					    int msg_type, double event_time)
{
  NodeInfo *node1, *node2;
  float now;

  // First we find the node
  node1 = findNode(node_id);
  if (node1 == NULL){
    fprintf(stderr, "Node %d cannot be found !\n", node_id);
    return;
  }

  // Find the second node, if any
  if (extra != -1){
    node2 = findNode(extra);
    if (node2 == NULL){
      fprintf(stderr, "Node %d cannot be found !\n", extra);
      return;
    }
  }
  else
    node2 = NULL;

  // Add nodes to nam, if necessary
  if (node1->initialized_ == false)
    addNode(node1);

  if (node2)
    if (node2->initialized_ == false)
      addNode(node2);

  now = getCurrentTime();

  if (sending){
    // Packet send event

    // Update node1's label
    printf("n -t %f -s %d -S DLABEL -l \"out:%d in:%d\" -L \"out:%d in:%d\"\n",
	   now, node1->node_id_, (node1->packets_sent_ + 1),
	   node1->packets_received_, node1->packets_sent_,
	   node1->packets_received_);
    fflush(NULL);

    // Update packet count
    node1->packets_sent_++;
    packets_out_++;

    // We just animate broadcast packets
    if (broadcast){
      printf("h -t %f -s %d -d -1 -p MAC -e 200000 -c %d -a %d -i 0 -k MAC\n",
	     now, node_id, node_id, msg_colors[msg_type]);
      fflush(NULL);
    }
  }
  else{
    // Packet receive event

    // Update node1's label
    printf("n -t %f -s %d -S DLABEL -l \"out:%d in:%d\" -L \"out:%d in:%d\"\n",
	   now, node1->node_id_, node1->packets_sent_,
	   (node1->packets_received_ + 1), node1->packets_sent_,
	   node1->packets_received_);
    fflush(NULL);

    // Update packet count
    node1->packets_received_++;
    packets_in_++;

    // Animate packet if not broadcast
    if (!broadcast){
      printf("h -t %f -s %d -d %d -p AODV -e 200000 -c %d -a %d -i 0 -k MAC\n",
	     now, extra, node_id, node_id, msg_colors[msg_type]);
      fflush(NULL);
    }
  }
}

void NamInterface::processEvent()
{
  TimerCallback *callback;
  double event_time;
  int msg_type = -1;
  bool sending, broadcast;
  char *found;
  int node_id;
  int extra;
  int len;

  // Check if nam_interface is initialized
  if (initialized_ == false){

    // Initialize nam_interface
    configNam();
    initialized_ = true;

    // Start sending dummy events
    sendDummyEvent();

    // Initialize label
    sendLabelEvent();

    // Add timers for sending dummy events and updating label
    callback = new SendDummyEventTimer(this);
    timers_manager_->addTimer(DUMMY_EVENT_PERIOD, callback);

    callback = new SendLabelEventTimer(this);
    timers_manager_->addTimer(LABEL_EVENT_PERIOD, callback);
  }

  // Look for a Diffusion Log Event
  found = strstr(&buf[0], "Diffusion Log");

  if (found){
    // This is a diffusion log event !

    found = strstr(&buf[0], "e");
    if (found == NULL){
      fprintf(stderr, "Cannot locate timestamp !\n");
      return;
    }

    // Advance until after "Time"
    found++;

    // Get timestamp
    event_time = atof(found);
    fprintf(stderr, "Time = %f ", event_time);

    // Get node ID
    found = strstr(found, "e");
    if (found == NULL){
      fprintf(stderr, "Cannot locate node ID !\n");
      return;
    }

    // Advance until after "Node"
    found++;

    // Get node id
    node_id = atoi(found);
    fprintf(stderr, "Node id = %d ", node_id);

    // Get sending/receive mode
    found = strstr(&buf[0], "sending");
    if (found){
      fprintf(stderr, "sending ");
      sending = true;
    }
    else{
      found = strstr(&buf[0], "received");
      if (found == NULL){
	fprintf(stderr, "Cannot read packet event !\n");
	return;
      }
      fprintf(stderr, "received ");
      sending = false;
    }

    // Get unicast/broadcast mode
    found = strstr(&buf[0], "broadcast");
    if (found){
      fprintf(stderr, "broadcast ");
      broadcast = true;
    }
    else{
      found = strstr(&buf[0], "unicast");
      if (found == NULL){
	fprintf(stderr, "Cannot read packet mode !\n");
	return;
      }
      fprintf(stderr, "unicast ");
      broadcast = false;
    }

    // Read packet length
    found = strstr(found, " ");

    if (found == NULL){
      fprintf(stderr, "Cannot read packet len !\n");
      return;
    }

    len = atoi(found);
    fprintf(stderr, "%d bytes ", len);

    // Read destination/source node
    found = strstr(found, "node");
    if (found == NULL){
      fprintf(stderr, "Cannot read extra info !\n");
      return;
    }

    found = strstr(found, " ");
    if (found == NULL){
      fprintf(stderr, "Cannot read extra info !\n");
      return;
    }

    extra = atoi(found);
    fprintf(stderr, "node %d ", extra);

    // Read message type
    found = strstr(found, "message ");
    if (found == NULL){
      fprintf(stderr, "Cannot read message type !\n");
      return;
    }

    found = strstr(found, " ");
    if (found == NULL){
      fprintf(stderr, "Cannot read message type !\n");
      return;
    }

    // Advance past the space
    found++;

    for (int i = 0; i < (sizeof(msg_types)/sizeof(msg_types[0])); i++){
      if (strstr(found, msg_types[i])){
	msg_type = i;
	break;
      }
    }

    if (msg_type < 0){
      fprintf(stderr, "Cannot read message type !\n");
      return;
    }

    fprintf(stderr, "message %s\n", msg_types[msg_type]);

    // Process Diffusion Log Event
    processDiffusionLogEvent(node_id, sending, broadcast, len, extra,
			     msg_type, event_time);

    return;
  }

  fprintf(stderr, "Unknown message received !\n");
  fprintf(stderr, "Message Received :%s\n", &buf[0]);
}

void NamInterface::run()
{
  struct timeval tv;
  int status;
  fd_set fds;

  while(1){
    // Add stdin to file descriptor set
    FD_ZERO(&fds);
    FD_SET(0, &fds);

    timers_manager_->nextTimerTime(&tv);
    if (tv.tv_sec == 0 && tv.tv_usec == 0){
      // Timer has expired !
      timers_manager_->executeNextTimer();
      continue;
    }

    // Check if there's information available
    status = select(1, &fds, NULL, NULL, &tv);

    if (status == 0){
      // We process all timers that have expired
      timers_manager_->nextTimerTime(&tv);
      while (tv.tv_sec == 0 && tv.tv_usec == 0){
	// Timer at the head of the queue has expired
	timers_manager_->executeNextTimer();
	timers_manager_->nextTimerTime(&tv);
      }
    }

    if (status > 0){
      cin.getline(buf, BUFFER_SIZE - 1);

      if (cin.eof())
	return;

      processEvent();
    }

    if (status < 0){
      fprintf(stderr, "Select returned %d\n", status);
    }
  }
}

void NamInterface::readConfigFile()
{
  NodeInfo *node;
  FILE *cfile;
  int flag;
  int id;
  int x, y;
  char color[50];
  char shape[50];

  cfile = fopen(config_file_, "r");

  // Check if we were able to open the config file
  if (cfile == NULL){
    fprintf(stderr, "Error: Can't open config file: %s !\n", config_file_);
    exit(1);
  }

  // Read node information from config file
  while (EOF != (flag = fscanf(cfile, "%d - %s - %s - %d - %d\n", &id,
			       color, shape, &x, &y))){
    if (flag != 5){
      fprintf(stderr, "Error in config file !\n");
      exit(1);
    }

    // Create a new node and add it to the node list
    node = new NodeInfo(id, x, y, &color[0], &shape[0]);
    nodes_.push_back(node);

    // Confirm node added to the screen
    fprintf(stderr, "Node %d added: Location: %d, %d, Color: %s, Shape: %s\n",
	    id, x, y, color, shape);
    fflush(NULL);
  }

  fclose(cfile);
}

NamInterface::NamInterface(int argc, char **argv)
{
  int opt;

  // Parse command line options
  size_x_ = DEFAULT_X;
  size_y_ = DEFAULT_Y;
  node_size_ = DEFAULT_NODE_SIZE;
  config_file_ = NULL;
  opterr = 0;

  while (1){
    opt = getopt(argc, argv, "hx:y:f:s:");

    switch (opt){

    case 'x':

      size_x_ = atoi(optarg);
      if (size_x_ < MIN_X || size_x_ > MAX_X){
	fprintf(stderr, "Error: Invalid X value !\n");
	usage(argv[0]);
      }

      break;

    case 'y':

      size_y_ = atoi(optarg);
      if (size_y_ < MIN_Y || size_y_ > MAX_Y){
	fprintf(stderr, "Error: Invalid Y value !\n");
	usage(argv[0]);
      }

      break;

    case 's':

      node_size_ = atoi(optarg);
      if (node_size_ < MIN_NODE_SIZE || node_size_ > MAX_NODE_SIZE){
	fprintf(stderr, "Error: Invalid node size !\n");
	usage(argv[0]);
      }

      break;

    case 'f':

      if (!strncasecmp(optarg, "-", 1)){
	fprintf(stderr, "Error: Parameter is missing !\n");
	usage(argv[0]);
      }

      config_file_ = strdup(optarg);

      break;

    case 'h':

      usage(argv[0]);

      break;

    case '?':

      fprintf(stderr,
	      "Error: %c isn't a valid option or its parameter is missing !\n",
	      optopt);
      usage(argv[0]);

      break;

    case ':':

      fprintf(stderr, "Error: Parameter missing !\n");
      usage(argv[0]);

      break;
    }

    if (opt == -1)
      break;
  }

  zero_.tv_sec = 0;
  zero_.tv_usec = 0;
  initialized_ = false;
  packets_in_ = 0;
  packets_out_ = 0;
  labels_ = 0;

  if (!config_file_)
    config_file_ = strdup("config.nam");

  // Initialize event queue and read config file
  timers_manager_ = new TimerManager;
  readConfigFile();
}

void NamInterface::usage(char *s)
{
  fprintf(stderr, "Usage: %s [-x max_x] [-y max_y] [-f file] [-s size] [-h]\n\n", s);
  fprintf(stderr, "\t-x - Sets max x coordinates in Nam 100-4000 (1200)\n");
  fprintf(stderr, "\t-y - Sets max y coordinates in Nam 100-4000 (1200)\n");
  fprintf(stderr, "\t-f - Uses alternate config file (config.nam)\n");
  fprintf(stderr, "\t-s - Sets node size 10-100 (50)\n");
  fprintf(stderr, "\t-h - Prints this information\n");
  fprintf(stderr, "\n");
  exit(0);
}

int main(int argc, char **argv)
{
  NamInterface *nam;

  nam = new NamInterface(argc, argv);
  nam->run();

  return 0;
}
